import contextlib
import platform

import qBitrr.logger  # noqa

with contextlib.suppress(ImportError):
    if platform.python_implementation() == "CPython":
        # Only replace complexjson on CPython
        # On PyPy it shows a SystemError when attempting to
        # decode the responses from qbittorrentapi
        import requests
        import ujson

        requests.models.complexjson = ujson



class qBitManagerError(Exception):
    """Base Exception"""


class UnhandledError(qBitManagerError):
    """Use to raise when there an unhandled edge case"""


class ConfigException(qBitManagerError):
    """Base Exception for Config related exceptions"""


class ArrManagerException(qBitManagerError):
    """Base Exception for Arr related Exceptions"""


class SkipException(qBitManagerError):
    """Dummy error to skip actions"""


class RequireConfigValue(qBitManagerError):
    """Exception raised when a config value requires a value."""

    def __init__(self, config_class: str, config_key: str):
        self.message = f"Config key '{config_key}' in '{config_class}' requires a value."


class NoConnectionrException(qBitManagerError):
    def __init__(self, message: str, type: str = "delay"):
        self.message = message
        self.type = type


class DelayLoopException(qBitManagerError):
    def __init__(self, length: int, type: str):
        self.type = type
        self.length = length


class RestartLoopException(ArrManagerException):
    """Exception to trigger a loop restart"""



from __future__ import annotations

import atexit
import itertools
import logging
import pathlib
import sys
import time
from multiprocessing import freeze_support

import pathos
import qbittorrentapi
import requests
from packaging import version as version_parser
from packaging.version import Version as VersionClass
from qbittorrentapi import APINames
from qbittorrentapi.decorators import login_required  # , response_text

from qBitrr.arss import ArrManager
from qBitrr.bundled_data import patched_version
from qBitrr.config import (
    APPDATA_FOLDER,
    CONFIG,
    CONFIG_EXISTS,
    ENABLE_LOGS,
    QBIT_DISABLED,
    SEARCH_ONLY,
    process_flags,
)
from qBitrr.env_config import ENVIRO_CONFIG
from qBitrr.ffprobe import FFprobeDownloader
from qBitrr.home_path import HOME_PATH
from qBitrr.logger import run_logs
from qBitrr.utils import ExpiringSet, absolute_file_paths

if CONFIG_EXISTS:
    from qBitrr.arss import ArrManager
else:
    sys.exit(0)

CHILD_PROCESSES = []

logger = logging.getLogger("qBitrr")
# if ENABLE_LOGS:
#     logs_folder = HOME_PATH.joinpath("logs")
#     logs_folder.mkdir(parents=True, exist_ok=True)
#     logs_folder.chmod(mode=0o777)
#     logfile = logs_folder.joinpath("qBitrr.log")
#     if pathlib.Path(logfile).is_file():
#         logold = logs_folder.joinpath( "qBitrr.log.old")
#         if pathlib.Path(logold).exists():
#             logold.unlink()
#         logfile.rename(logold)
#     fh = logging.FileHandler(logfile)
#     logger.addHandler(fh)
run_logs(logger)


class qBitManager:
    min_supported_version = VersionClass("4.3.9")
    soft_not_supported_supported_version = VersionClass("4.4.4")
    max_supported_version = VersionClass("4.6.7")
    _head_less_mode = False

    def __init__(self):
        self._name = "Manager"
        self.qBit_Host = CONFIG.get("qBit.Host", fallback="localhost")
        self.qBit_Port = CONFIG.get("qBit.Port", fallback=8105)
        self.qBit_UserName = CONFIG.get("qBit.UserName", fallback=None)
        self.qBit_Password = CONFIG.get("qBit.Password", fallback=None)
        self.qBit_v5 = CONFIG.get("qBit.v5", fallback=False)
        self.logger = logging.getLogger(f"qBitrr.{self._name}")
        if ENABLE_LOGS:
            logs_folder = HOME_PATH.joinpath("logs")
            logs_folder.mkdir(parents=True, exist_ok=True)
            logs_folder.chmod(mode=0o777)
            logfile = logs_folder.joinpath(self._name + ".log")
            if pathlib.Path(logfile).is_file():
                logold = logs_folder.joinpath(self._name + ".log.old")
                if pathlib.Path(logold).exists():
                    logold.unlink()
                logfile.rename(logold)
            fh = logging.FileHandler(logfile)
            self.logger.addHandler(fh)
        run_logs(self.logger)
        self.logger.debug(
            "qBitTorrent Config: Host: %s Port: %s, Username: %s, Password: %s",
            self.qBit_Host,
            self.qBit_Port,
            self.qBit_UserName,
            self.qBit_Password,
        )
        self._validated_version = False
        self.client = None
        self.current_qbit_version = None
        if not any([QBIT_DISABLED, SEARCH_ONLY]):
            self.client = qbittorrentapi.Client(
                host=self.qBit_Host,
                port=self.qBit_Port,
                username=self.qBit_UserName,
                password=self.qBit_Password,
                SIMPLE_RESPONSES=False,
            )
            try:
                self.current_qbit_version = version_parser.parse(self.client.app_version())
                self._validated_version = True
            except BaseException:
                self.current_qbit_version = self.min_supported_version
                self.logger.error(
                    "Could not establish qBitTorrent version, "
                    "you may experience errors, please report this error."
                )
            self._version_validator()
        self.expiring_bool = ExpiringSet(max_age_seconds=10)
        self.cache = {}
        self.name_cache = {}
        self.should_delay_torrent_scan = False  # If true torrent scan is delayed by 5 minutes.
        self.child_processes = []
        self.ffprobe_downloader = FFprobeDownloader()
        try:
            if not any([QBIT_DISABLED, SEARCH_ONLY]):
                self.ffprobe_downloader.update()
        except Exception as e:
            self.logger.error(
                "FFprobe manager error: %s while attempting to download/update FFprobe", e
            )
        self.arr_manager = ArrManager(self).build_arr_instances()
        run_logs(self.logger)

    def _version_validator(self):
        validated = False
        if self.qBit_v5:
            if self.min_supported_version <= self.current_qbit_version:
                validated = True
        else:
            if (
                self.min_supported_version
                <= self.current_qbit_version
                <= self.max_supported_version
            ):
                validated = True

        if self._validated_version and validated:
            self.logger.info(
                "Current qBitTorrent version is supported: %s", self.current_qbit_version
            )
        elif not self._validated_version and validated:
            self.logger.warning(
                "Could not validate current qBitTorrent version, assuming: %s",
                self.current_qbit_version,
            )
            time.sleep(10)
        else:
            self.logger.critical(
                "You are currently running qBitTorrent version %s, "
                "Supported version range is %s to < %s",
                self.current_qbit_version,
                self.min_supported_version,
                self.max_supported_version,
            )
            sys.exit(1)

    # @response_text(str)
    @login_required
    def app_version(self, **kwargs):
        return self.client._get(
            _name=APINames.Application,
            _method="version",
            _retries=0,
            _retry_backoff_factor=0,
            **kwargs,
        )

    @property
    def is_alive(self) -> bool:
        try:
            if 1 in self.expiring_bool or self.client is None:
                return True
            self.client.app_version()
            self.logger.trace("Successfully connected to %s:%s", self.qBit_Host, self.qBit_Port)
            self.expiring_bool.add(1)
            return True
        except requests.RequestException:
            self.logger.warning("Could not connect to %s:%s", self.qBit_Host, self.qBit_Port)
        self.should_delay_torrent_scan = True
        return False

    def get_child_processes(self) -> list[pathos.helpers.mp.Process]:
        run_logs(self.logger)
        self.logger.debug("Managing %s categories", len(self.arr_manager.managed_objects))
        count = 0
        procs = []
        for arr in self.arr_manager.managed_objects.values():
            numb, processes = arr.spawn_child_processes()
            count += numb
            procs.extend(processes)
        return procs

    def run(self):
        try:
            self.logger.debug("Starting %s child processes", len(self.child_processes))
            [p.start() for p in self.child_processes]
            [p.join() for p in self.child_processes]
        except KeyboardInterrupt:
            self.logger.info("Detected Ctrl+C - Terminating process")
            sys.exit(0)
        except BaseException as e:
            self.logger.info("Detected Ctrl+C - Terminating process: %r", e)
            sys.exit(1)


def run():
    global CHILD_PROCESSES
    early_exit = process_flags()
    if early_exit is True:
        sys.exit(0)
    logger.info("Starting qBitrr: Version: %s.", patched_version)
    try:
        manager = qBitManager()
    except NameError:
        sys.exit(0)
    run_logs(logger)
    logger.debug("Environment variables: %r", ENVIRO_CONFIG)
    try:
        if CHILD_PROCESSES := manager.get_child_processes():
            manager.run()
        else:
            logger.warning(
                "No tasks to perform, if this is unintended double check your config file."
            )
    except KeyboardInterrupt:
        logger.info("Detected Ctrl+C - Terminating process")
        sys.exit(0)
    except Exception:
        logger.info("Attempting to terminate child processes, please wait a moment.")
        for child in manager.child_processes:
            child.kill()


def cleanup():
    for p in CHILD_PROCESSES:
        p.kill()
        p.terminate()


def file_cleanup():
    extensions = [".db", ".db-shm", ".db-wal"]
    all_files_in_folder = list(absolute_file_paths(APPDATA_FOLDER))
    for file, ext in itertools.product(all_files_in_folder, extensions):
        if file.name.endswith(ext):
            APPDATA_FOLDER.joinpath(file).unlink(missing_ok=True)


atexit.register(cleanup)


if __name__ == "__main__":
    freeze_support()
    file_cleanup()
    run()



from __future__ import annotations

import argparse
import contextlib
import pathlib
import shutil
import sys

from qBitrr.bundled_data import license_text, patched_version
from qBitrr.env_config import ENVIRO_CONFIG
from qBitrr.gen_config import MyConfig, _write_config_file, generate_doc
from qBitrr.home_path import APPDATA_FOLDER, HOME_PATH


def process_flags() -> argparse.Namespace | bool:
    parser = argparse.ArgumentParser(description="An interface to interact with qBit and *arrs.")
    parser.add_argument(
        "--gen-config",
        "-gc",
        dest="gen_config",
        help="Generate a config file in the current working directory",
        action="store_true",
    )
    parser.add_argument(
        "-v", "--version", action="version", version=f"qBitrr version: {patched_version}"
    )

    parser.add_argument(
        "-l",
        "--license",
        dest="license",
        action="store_const",
        const=license_text,
        help="Show the qBitrr's licence",
    )
    parser.add_argument(
        "-s",
        "--source",
        action="store_const",
        dest="source",
        const="Source code can be found on: https://github.com/Feramance/qBitrr",
        help="Shows a link to qBitrr's source",
    )

    args = parser.parse_args()

    if args.gen_config:
        from qBitrr.gen_config import _write_config_file

        _write_config_file()
        return True
    elif args.license:
        print(args.license)
        return True
    elif args.source:
        print(args.source)
        return True
    return args


COPIED_TO_NEW_DIR = False
file = "config.toml"
CONFIG_EXISTS = True
CONFIG_FILE = HOME_PATH.joinpath(file)
CONFIG_PATH = pathlib.Path(f"./{file}")
if any(
    a in sys.argv
    for a in [
        "--gen-config",
        "-gc",
        "--version",
        "-v",
        "--license",
        "-l",
        "--source",
        "-s",
        "-h",
        "--help",
    ]
):
    CONFIG = MyConfig(CONFIG_FILE, config=generate_doc())
    COPIED_TO_NEW_DIR = None
elif (not CONFIG_FILE.exists()) and (not CONFIG_PATH.exists()):
    print(f"{file} has not been found")

    CONFIG_FILE = _write_config_file(docker=True)
    print(f"'{CONFIG_FILE.name}' has been generated")
    print('Rename it to "config.toml" then edit it and restart the container')

    CONFIG_EXISTS = False

elif CONFIG_FILE.exists():
    CONFIG = MyConfig(CONFIG_FILE)
else:
    with contextlib.suppress(
        Exception
    ):  # If file already exist or can't copy to APPDATA_FOLDER ignore the exception
        shutil.copy(CONFIG_PATH, CONFIG_FILE)
        COPIED_TO_NEW_DIR = True
    CONFIG = MyConfig("./config.toml")

if COPIED_TO_NEW_DIR is not None:
    # print(f"STARTING QBITRR | {CONFIG.path} |\n{CONFIG}")
    print("STARTING QBITRR")
else:
    print(f"STARTING QBITRR |  CONFIG_FILE={CONFIG_FILE} | CONFIG_PATH={CONFIG_PATH}")

FFPROBE_AUTO_UPDATE = (
    CONFIG.get("Settings.FFprobeAutoUpdate", fallback=True)
    if ENVIRO_CONFIG.settings.ffprobe_auto_update is None
    else ENVIRO_CONFIG.settings.ffprobe_auto_update
)
FAILED_CATEGORY = ENVIRO_CONFIG.settings.failed_category or CONFIG.get(
    "Settings.FailedCategory", fallback="failed"
)
RECHECK_CATEGORY = ENVIRO_CONFIG.settings.recheck_category or CONFIG.get(
    "Settings.RecheckCategory", fallback="recheck"
)
TAGLESS = ENVIRO_CONFIG.settings.tagless or CONFIG.get("Settings.Tagless", fallback=False)
CONSOLE_LOGGING_LEVEL_STRING = ENVIRO_CONFIG.settings.console_level or CONFIG.get(
    "Settings.ConsoleLevel", fallback="INFO"
)
ENABLE_LOGS = ENVIRO_CONFIG.settings.logging or CONFIG.get("Settings.Logging", fallback=True)
COMPLETED_DOWNLOAD_FOLDER = (
    ENVIRO_CONFIG.settings.completed_download_folder
    or CONFIG.get_or_raise("Settings.CompletedDownloadFolder")
)
FREE_SPACE = ENVIRO_CONFIG.settings.free_space or CONFIG.get("Settings.FreeSpace", fallback="-1")
NO_INTERNET_SLEEP_TIMER = ENVIRO_CONFIG.settings.no_internet_sleep_timer or CONFIG.get(
    "Settings.NoInternetSleepTimer", fallback=60
)
LOOP_SLEEP_TIMER = ENVIRO_CONFIG.settings.loop_sleep_timer or CONFIG.get(
    "Settings.LoopSleepTimer", fallback=5
)
SEARCH_LOOP_DELAY = ENVIRO_CONFIG.settings.search_loop_delay or CONFIG.get(
    "Settings.SearchLoopDelay", fallback=-1
)
AUTO_PAUSE_RESUME = ENVIRO_CONFIG.settings.auto_pause_resume or CONFIG.get(
    "Settings.AutoPauseResume", fallback=True
)
PING_URLS = ENVIRO_CONFIG.settings.ping_urls or CONFIG.get(
    "Settings.PingURLS", fallback=["one.one.one.one", "dns.google.com"]
)
IGNORE_TORRENTS_YOUNGER_THAN = ENVIRO_CONFIG.settings.ignore_torrents_younger_than or CONFIG.get(
    "Settings.IgnoreTorrentsYoungerThan", fallback=600
)
QBIT_DISABLED = (
    CONFIG.get("qBit.Disabled", fallback=False)
    if ENVIRO_CONFIG.qbit.disabled is None
    else ENVIRO_CONFIG.qbit.disabled
)
SEARCH_ONLY = ENVIRO_CONFIG.overrides.search_only
PROCESS_ONLY = ENVIRO_CONFIG.overrides.processing_only

if QBIT_DISABLED and PROCESS_ONLY:
    print("qBittorrent is disabled yet QBITRR_OVERRIDES_PROCESSING_ONLY is enabled")
    print(
        "Processing monitors qBitTorrents downloads "
        "therefore it depends on a health qBitTorrent connection"
    )
    print("Exiting...")
    sys.exit(1)

if SEARCH_ONLY and QBIT_DISABLED is False:
    QBIT_DISABLED = True
    print("QBITRR_OVERRIDES_SEARCH_ONLY is enabled, forcing qBitTorrent setting off")

# Settings Config Values
FF_VERSION = APPDATA_FOLDER.joinpath("ffprobe_info.json")
FF_PROBE = APPDATA_FOLDER.joinpath("ffprobe")



from distutils.util import strtobool
from typing import Optional

import environ


class Converter:
    @staticmethod
    def int(value: Optional[str]) -> Optional[int]:
        return None if value is None else int(value)

    @staticmethod
    def list(value: Optional[str], delimiter=",", converter=str) -> Optional[list]:
        return None if value is None else list(map(converter, value.split(delimiter)))

    @staticmethod
    def bool(value: Optional[str]) -> Optional[bool]:
        return None if value is None else strtobool(value) == 1


@environ.config(prefix="QBITRR", frozen=True)
class AppConfig:
    @environ.config(prefix="OVERRIDES", frozen=True)
    class Overrides:
        search_only = environ.var(None, converter=Converter.bool)
        processing_only = environ.var(None, converter=Converter.bool)
        data_path = environ.var(None)

    @environ.config(prefix="SETTINGS", frozen=True)
    class Settings:
        console_level = environ.var(None)
        logging = environ.var(None, converter=Converter.bool)
        completed_download_folder = environ.var(None)
        free_space = environ.var(None)
        no_internet_sleep_timer = environ.var(None, converter=Converter.int)
        loop_sleep_timer = environ.var(None, converter=Converter.int)
        search_loop_delay = environ.var(None, converter=Converter.int)
        auto_pause_resume = environ.var(None, converter=Converter.bool)
        failed_category = environ.var(None)
        recheck_category = environ.var(None)
        tagless = environ.var(None, converter=Converter.bool)
        ignore_torrents_younger_than = environ.var(None, converter=Converter.int)
        ping_urls = environ.var(None, converter=Converter.list)
        ffprobe_auto_update = environ.var(None, converter=Converter.bool)

    @environ.config(prefix="QBIT", frozen=True)
    class qBit:
        disabled = environ.var(None, converter=Converter.bool)
        host = environ.var(None)
        port = environ.var(None, converter=Converter.int)
        username = environ.var(None)
        password = environ.var(None)
        v5 = environ.var(False)

    overrides: Overrides = environ.group(Overrides)
    settings: Settings = environ.group(Settings)
    qbit: qBit = environ.group(qBit)


ENVIRO_CONFIG: AppConfig = environ.to_config(AppConfig)
